/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* bcpoly.c: manipulate rational expressions for bc */

#include "eclib.h"

/* substitute in a polynomial function */
struct POLY *polysub(a, list) /*###*/
struct POLY *a;
long list;
{
register i;
struct POLY *accum, *p, *q, *v;
long mask;
char rem_exp;
struct TERM *t;

/* always substitute variables that are roots of other things */
for(i=0; i<26; ++i) if(vbr[i] > 1) list |= (1L<<i);

while(1) { /* keep substituting */

/* build a mask of extant variables */
mask = 0L;
for(t=a->terms; t<a->terms + a->nterms; ++t)
for(i=0; i<t->nvars; ++i)
if(t->exps[i] >= vbr[t->vars[i]])
mask |= 1L << t->vars[i];

mask &= list;
if(!mask) break; /* done */

accum = &pzero;
for(t=a->terms; t<a->terms + a->nterms; ++t) {

p = newpoly(1);
p->terms[0].coef = iadd(t->coef, &izero);

for(i=0; i<t->nvars; ++i) {
rem_exp = t->exps[i]; /* remaining exponent */
if(mask & (1L << t->vars[i])) { /* substitute */
digit.value[0] = t->exps[i]/vbr[t->vars[i]];
rem_exp = t->exps[i]%vbr[t->vars[i]];
if(digit.value[0] > 0) {
v = rexp(vb[t->vars[i]], &digit);
q = rmul(p, v);
if(digit.value[0] > 1) polyfree(v);
polyfree(p);
p = q;
}
}
if(rem_exp) {
pvar.terms[0].vars[0] = t->vars[i];
pvar.terms[0].degree = pvar.terms[0].exps[0] = rem_exp;
q = rmul(&pvar, p);
polyfree(p);
p = q;
}
} /* end loop over variables in the term */

/* p is built, add it to the accumulator */
q = radd(accum, p);
bcfree(p);
bcfree(accum);
accum = q;
} /* end loop over terms */

a = accum;
}

return a;
} /* polysub */

/* sort and combine terms in a polynomial */
struct POLY *polysimp(p, list) /*###*/
struct POLY *p;
long list;
{
register struct TERM *t;
struct POLY *q;
register j;
char e; /* exponent on i^e */
char dir;
char swap_c;
struct TERM swap_t;
int changed;

top:
if(p == &pzero) return p; /* kludge */
/* this isn't just an optimization, dropping 0 from pzero is a disaster */

/* don't allocate anything new, just sort terms in line */
/* simple bubble sort */

/* first sort the variables in the terms */
for(t=p->terms; t<p->terms + p->nterms; ++t) {
do {
changed = 0;
for(j=0; j<t->nvars-1; ++j) {
dir = t->vars[j] - t->vars[j+1];
if(dir < 0) continue;
if(dir > 0) {
swap_c = t->exps[j], t->exps[j] = t->exps[j+1], t->exps[j+1] = swap_c;
swap_c = t->vars[j], t->vars[j] = t->vars[j+1], t->vars[j+1] = swap_c;
changed = 1;
continue;
}

/* combine variables */
t->exps[j] += t->exps[j+1];
if(t->exps[j] & 0x80) { eprint(er_exp); t->exps[j] = 1; }
--t->nvars;
t->vars[j+1] = t->vars[t->nvars];
t->exps[j+1] = t->exps[t->nvars];
--j; /* look again */
} /* end loop over variables */

} while(changed); /* end bubble sort */
} /* end loop over terms */

/* now sort the terms */
do {
changed = 0;

for(t=p->terms; t<p->terms + p->nterms - 1; ++t) {
dir = t[0].degree - t[1].degree;
if(!dir)
dir = t[1].nvars - t[0].nvars;
if(!dir)
for(j=0; j<t[0].nvars; ++j) {
if(dir = t[1].vars[j] - t[0].vars[j]) break;
if(dir = t[0].exps[j] - t[1].exps[j]) break;
}
if(dir > 0) continue;
if(dir < 0) {
swap_t = t[0], t[0] = t[1], t[1] = swap_t;
changed = 1;
continue;
}

/* combine like terms */
t[0].coef = iadd(t[0].coef, t[1].coef);
--p->nterms;
t[1] = p->terms[p->nterms];
--t; /* look again */
} /* end loop over terms */
} while(changed); /* end bubble sort */

/* now drop terms with 0 coefficients */
for(t=p->terms; t<p->terms + p->nterms; ++t) {
struct TERM *t1;
if(t->coef->size) continue; /* not 0 */
--p->nterms;
for(t1=t; t1<p->terms + p->nterms; ++t1)
t1[0] = t1[1];
--t; /* look again */
} /* end loop over terms */

/* can't have an empty polynomial */
if(p->nterms == 0)
return &pzero;

q = polysub(p, list);
if(q == p) return q;
p = q;
goto top;
} /* polysimp */

/* reduce a polynomial, divide through by gcd of coefficients */
reduce(p) /*###*/
struct POLY *p;
{
register i;
unsigned n;
char first;
struct VINT *g, *temp;

if(!p->next) return; /* no denominator, nothing to reduce */

g = p->terms[0].coef;
first = 1;
for(i=1; i<p->nterms; ++i) {
temp = igcd(g, p->terms[i].coef);
if(!first) bcfree(g);
g = temp;
first = 0;
if(g->size == 1 && g->value[0] == 1) goto dencheck;
}

for(i=0; i<p->next->nterms; ++i) {
temp = igcd(g, p->next->terms[i].coef);
if(!first) bcfree(g);
g = temp;
first = 0;
if(g->size == 1 && g->value[0] == 1) goto dencheck;
}

for(i=0; i<p->nterms; ++i) {
temp = idiv(p->terms[i].coef, g);
bcfree(p->terms[i].coef);
p->terms[i].coef = temp;
}

for(i=0; i<p->next->nterms; ++i) {
temp = idiv(p->next->terms[i].coef, g);
bcfree(p->next->terms[i].coef);
p->next->terms[i].coef = temp;
}

dencheck: /* check negative denominator */
if(!first) bcfree(g);
if(p->next->nterms == 1) {
if(p->next->terms[0].coef->size & 0x8000) {
for(i=0; i<p->nterms; ++i) {
temp = isub(&izero, p->terms[i].coef);
bcfree(p->terms[i].coef);
p->terms[i].coef = temp;
}
temp = isub(&izero, p->next->terms[0].coef);
bcfree(p->next->terms[0].coef);
p->next->terms[0].coef = temp;
} /* end negative denominator */

/* 1 in the denominator can be dropped,
 * but I decided not to do it.
 * When I'm working in fractions, sometimes I want to stay in fractions.
if(p->next->terms[0].coef->size == 1 && p->next->terms[0].coef->value[0] == 1
&& p->next->terms[0].nvars == 0)
polyfree(p->next), p->next = 0;
 */
} /* end single term in the denominator */

/* 0 numerator */
if(p->terms[0].coef->size == 0)
polyfree(p->next), p->next = 0;

return;
} /* reduce */


/* polynomial arithmetic */

/* polynomial add */
struct POLY *padd(a, b) /*###*/
struct POLY *a, *b;
{
register struct POLY *p;
int i, n;

/* just concatenate, the above routine will clean things up */
n = a->nterms + b->nterms;
p = newpoly(n);
p->nterms = n;
p->next = 0;
for(i=0; i<a->nterms; ++i)
p->terms[i] = a->terms[i];
n = i;
for(i=0; i<b->nterms; ++i,++n)
p->terms[n] = b->terms[i];
return polysimp(p, 0L);
}

/* polynomial multiply */
struct POLY *pmul(a, b) /*###*/
struct POLY *a, *b;
{
register struct POLY *p;
int i, j, n;

n = a->nterms * b->nterms;
p = newpoly(n);

/* outer product */
n = 0;
for(i=0; i<a->nterms; ++i)
for(j=0; j<b->nterms; ++j)
{
register k, l;
int m;
struct TERM *a1, *b1, *t;

a1 = &a->terms[i];
b1 = &b->terms[j];
t = &p->terms[n];
k = l = m = 0;

/* merge variables together */
while(k < a1->nvars || l < b1->nvars)
{
if(m == MAXVAR_TERM) { eprint("too many variables in a term"); return &pzero; }
if(l == b1->nvars) goto add1;
if(k == a1->nvars) goto add2;
if(a1->vars[k] == b1->vars[l])
{
t->vars[m] = a1->vars[k];
t->exps[m] = a1->exps[k] + b1->exps[l];
if(t->exps[m] & 0x80) { eprint(er_exp); return &pzero; }
++k, ++l, ++m;
continue;
}
if(b1->vars[l] < a1->vars[k]) goto add2;
add1:
t->vars[m] = a1->vars[k];
t->exps[m] = a1->exps[k];
++k, ++m;
continue;
add2:
t->vars[m] = b1->vars[l];
t->exps[m] = b1->exps[l];
++l, ++m;
continue;
} /* end merging variables */

t->coef = imul(a1->coef, b1->coef);
t->nvars = m;
t->degree = a1->degree + b1->degree;
++n;
} /* end double loop over terms */

return polysimp(p, 0L);
}


/* rational functions */

/* rational exponentiation */
struct POLY *rexp(a, e) /*###*/
struct POLY *a;
struct VINT *e;
{
struct POLY *p, *q;
short n;

/* constant */
if(isconst(a))
{
p = newpoly(1);
p->terms[0].coef = vexp(a->terms[0].coef, e);
return p;
}

n = e->size;
n &= 0x7fff;
if(n > 1) { eprint(er_exp); return &pzero; }

if(n == 0) return &pone;
n = e->value[0];

p = a;
while(--n)
{
q = rmul(p, a);
if(p != a) polyfree(p); /* free intermediates */
p = q;
}

if(e->size < 0)
{ /* flip it */
q = rdiv(&pone, p);
if(p != a) polyfree(p); /* free intermediates */
p = q;
}

return p;
}


/* rational add */
struct POLY *radd(a, b) /*###*/
struct POLY *a, *b;
{
struct POLY *p, *p1, *p2;

if(b->next && !a->next) p = a, a = b, b = p;

if(!a->next) return padd(a, b);

p2 = pmul(b, a->next);
if(b->next)
{
p1 = pmul(a, b->next);
p = padd(p1, p2);
bcfree(p1);
p->next = pmul(a->next, b->next);
} else {
p = padd(p2, a);
p->next = a->next; /* why not */
}
bcfree(p2);

return p;
}

/* rational subtract */
struct POLY *rsub(a, b) /*###*/
struct POLY *a, *b;
{
static struct VINT minus1 = {(short)0x8001, 0, {1}};
struct POLY *q;

pconst.terms[0].coef = &minus1;
q = rmul(&pconst, b);
a = radd(a, q);
bcfree(q);
return a;
}

/* rational multiply */
struct POLY *rmul(a, b) /*###*/
struct POLY *a, *b;
{
struct POLY *p;

p = pmul(a, b);
if(a->next)
{
if(b->next)
p->next = pmul(a->next, b->next);
else
p->next = a->next;
} else {
if(b->next) p->next = b->next;
}
return p;
}

/* rational divide */
struct POLY *rdiv(a, b) /*###*/
struct POLY *a, *b;
{
struct POLY *p;

/* constants? */
if(isconst(a) && isconst(b) && !intdiv)
{
p = newpoly(1);
p->terms[0].coef = idiv(a->terms[0].coef, b->terms[0].coef);
return p;
}

if(b->next) p = pmul(a, b->next);
else p = padd(a, &pzero);
if(a->next) p->next = pmul(b, a->next);
else p->next = padd(b, &pzero);

return p;
}

/* rational remainder */
struct POLY *rrem(a, b) /*###*/
struct POLY *a, *b;
{
struct POLY *p;

/* constants? */
if(a->nterms + b->nterms == 2)
{
p = newpoly(1);
p->terms[0].coef = irem(a->terms[0].coef, b->terms[0].coef);
return p;
}

return &pzero;
}

/* rational binary coefficient, x*(x-1)*(x-2)* ... (x-k+1) */
struct POLY *rchoose(a, k) /*###*/
struct POLY *a;
struct VINT *k;
{
struct POLY *p;

/* temporary */
if(!isconst(a)) return &pzero;

p = newpoly(1);
p->terms[0].coef = vchoose(a->terms[0].coef, k);
return p;
}

/* substitute in a rational function */
struct POLY *subst(a, list) /*###*/
struct POLY *a;
long list;
{
struct POLY *p, *q, *r, *save_d;

p = polysub(a, list);
if(a->next)
{
q = polysub(a->next, list);
/* This is subtile (a euphimism for kludged).
 * Polysub may not change a, simply passing it back as p, so that
 * p has a denominator.  One would normally just divide p by q and be
 * done with it, but (p=a) may still have its old denominator around
 * if there was nothing in the numerator to substitute.
 * We have to kill the denominator long enough to run the divide,
 * then put it back.  Sometimes I think this whole mess should
 * be rewritten. */
save_d = p->next;
p->next = 0;
r = rdiv(p, q);
p->next = save_d;
polyfree(p);
polyfree(q);
p = r;
}
return p;
}

/* assign a polynomial to a variable */
assign(var, p, exp) /*###*/
struct POLY *p, *exp;
{
register i;
char v, w; /* variables */
long list; /* keep list for recursion check */
long oldlist;
struct POLY *q;
struct TERM *t;
struct VINT *expint;
struct POLY *oldvar; /* save state */
char oldroot;

if(!exp) exp = &pone;
expint = exp->terms[0].coef;
if(!isconst(exp)) { eprint(er_lhs); return; }
if(expint->size > 1 || expint->value[0] > 127) {
eprint(er_exp);
return;
}
oldvar = vb[var];
oldroot = vbr[var];
vb[var] = p;
vbr[var] = expint->value[0];

/* ok, test for recursion */
list = (1L<<var);

do {
oldlist = list;

for(v=0; v<26; ++v) {
if(!(list & (1L<<v))) continue;
q = vb[v];
for(t=q->terms; t<q->terms + q->nterms; ++t)
for(i=0; i<t->nvars; ++i)
{
w = t->vars[i];
if(w == var && t->exps[i] >= vbr[var]) {
eprint("variable recursion in assignment");
vb[var] = oldvar; /* put it back */
vbr[var] = oldroot;
return;
}
list |= (1L<<w);
} /* end loop over terms in polynomial vb[v] */
} /* end loop over accessible variables */
} while(list != oldlist);

reduce(vb[var]);
return; /* ok */
} /* assign */

/* display polynomial */
/* stores it in dotval */
pdisp(p)
struct POLY *p;
{
register i;
struct TERM *t;
char first_v, first_t, first_p;
short minus;

reduce(p);
dotval = p;
first_p = 1;

for(; p; p=p->next) {
if(first_p) first_p = 0; else printf("  /  ");

first_t = 1;
for(t=p->terms; t<p->terms + p->nterms; ++t)
{

/* constant term might be 0 */
if((minus = t->coef->size) == 0 && p->nterms > 1) continue;
if(minus < 0) t->coef->size &= 0x7fff; /* i'll put it back, honest */
if(first_t) {
first_t = 0;
if(minus < 0) printf("- ");
} else {
if(minus < 0) printf(" - ");
else printf(" + ");
}

first_v = 1;
if(t->coef->size != 1 || t->coef->value[0] != 1 || t->nvars == 0) {
idisp(t->coef);
first_v = 0;
}
if(minus < 0) t->coef->size |= 0x8000; /* told ya */

for(i=0; i<t->nvars; ++i) {
if(first_v) first_v = 0; else printf("*");
printf("%c",t->vars[i] + 'a');
if(t->exps[i] > 1) printf("^%d", t->exps[i]);
} /* end loop printing variables */
} /* end loop printing terms */
} /* end loop printing polynomials in a chain */
} /* pdisp */

