/***********************************************************************
|  Genealogy management system
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* serial.c: routines for serial index lists */

#include <errno.h>

#include "eclib.h"

#undef NULL
#define NULL (struct serial_inmem *) 0

#define PATHLEN 140
#define FILELEN 14

static char tupbuf[TUPLEN]; /* large enough to hold any tuple */
static char logtupbuf[TUPLEN+8]; /* large enough to hold any tuple +8 */
static char badbuf[] = "@cannot use serial-internal buffer as an argument to insupd() routine";
char datadir[PATHLEN], indexdir[PATHLEN], logdir[PATHLEN], lockdir[PATHLEN];
static char datadir_len, indexdir_len, logdir_len, lockdir_len;
/* serial flags */
static char sf_dirs, sf_multi, sf_log;
static long ks[2];

serial_dirs(dd, id, ld, kd)
char *dd, *id, *ld, *kd; /* the directories */
{
extern int uncache_fd();

if(sf_dirs)
eclib_error(NODE_SERIAL_DIRS_2,0);
ZC_fn_uncache = uncache_fd;

dirsetup(datadir, dd, &datadir_len);
dirsetup(indexdir, id, &indexdir_len);
strcat(indexdir, "index_x");
indexdir_len += 6;
dirsetup(logdir, ld, &logdir_len);
strcat(logdir, "logfile");
dirsetup(lockdir, kd, &lockdir_len);
strcat(lockdir, "lock_x");
lockdir_len += 5;
idxval_read();
sf_dirs = 1;
} /* serial_dirs */

static dirsetup(dest, src, len_ptr)
char *dest, *src;
char *len_ptr;
{
int l = strlen(src);
if(l > PATHLEN-1-FILELEN-1)
eclib_error(SCAN_LONGNAME, PATHLEN-1-FILELEN-1);
strcpy(dest, src);
if(l) { strcat(dest, "/"); ++l; }
*len_ptr = l;
} /* dirsetup */

/* set and clear lock files, exclussive changes to the database */
lock_set(relnum)
{
int i, fd;
if(!sf_multi) return;
if(struct_desc[relnum].holdopen) return; /* happens during casc_upd_del() */
lockdir[lockdir_len] = 'a'+relnum;
for(i=0; i<15; ++i) {
fd = open(lockdir, O_BINARY|O_CREAT|O_EXCL, 0666);
if(fd >= 0) { /* ok */
close(fd);
return;
}
if(errno == EMFILE) { uncache_fd(); continue; }
if(errno != EEXIST)
eclib_error(NODE_MKLOCK, relnum);
#ifndef MSDOS
sleep(1);
#endif
}
eclib_error(NODE_LOCKED, relnum);
} /* lock_set */

lock_clear(relnum)
{
if(!sf_multi) return;
if(struct_desc[relnum].holdopen) return; /* happens during casc_upd_del() */
lockdir[lockdir_len] = 'a'+relnum;
unlink(lockdir);
} /* lock_clear */

static long file_manage(x, relnum, zero_ok, writing, id)
struct serial_inmem *x;
short relnum;
long id;
{
if(x) {
if(x->setup != 0x5a)
eclib_error(NODE_SERIAL_LIST_0, 0);
relnum = x->relnum;
}

if(struct_desc[relnum].nodetype != 1)
eclib_error(NODE_STRUCT_TYPE, relnum);
if(id) id -= x ? x->base : struct_desc[relnum].base;
if(id+zero_ok <= 0)
eclib_error(NODE_SERIAL_RANGE, relnum);
if(writing && struct_desc[relnum].rdonly)
eclib_error(NODE_RDONLY, relnum);
if(struct_desc[relnum].size > TUPLEN)
eclib_error(NODE_TOO_LARGE, TUPLEN);

if(!x) fd_setup(relnum);

return id;
} /* file_manage */

/* select a row.
 * No error occurs if you select a deleted or unused row,
 * but you get a null pointer back.
 * next_scan() skips these rows.
 * Make sure legitimate data cannot begin with AVAIL_CONSTANT. */
char *ZC_select_row_mem(x, id, dest)
struct serial_inmem *x;
long id;
char *dest;
{
short relnum, length;
char *s;

id = file_manage(x, 0, 0, 0, id);
relnum = x->relnum;
length = struct_desc[relnum].size;

if(x->n_nodes <= id)
return 0;
s = x->inmem + id * length;
if(*((long *)s) == AVAIL_CONSTANT) return 0;
if(dest) { memcpy(dest, s, length); s = dest; }
return s;
} /* ZC_select_row_mem */

char *ZC_select_row_file(relnum, id, dest)
short relnum;
long id;
char *dest;
{
short length, fd;

id = file_manage(NULL, relnum, 0, 0, id);
length = struct_desc[relnum].size;
fd = struct_desc[relnum].fd;
if(!dest) dest = tupbuf;

elseek(fd, id*length, 0);
if(!eread(fd, dest, length) ||
*((long *)dest) == AVAIL_CONSTANT)
dest = 0;

return dest;
} /* ZC_select_row_file */

ZC_delete_row_mem(x, id)
struct serial_inmem *x;
long id;
{
short relnum, length;
char *s;
long *lptr, *lptr_s;

id = file_manage(x, 0, 0, 1, id);
relnum = x->relnum;
length = struct_desc[relnum].size;

if(x->n_nodes <= id)
eclib_error(NODE_SERIAL_RANGE, relnum);
s = x->inmem + id * length;
lptr = (long *) (x->inmem);
lptr_s = (long *)s;
if(lptr_s[0] == AVAIL_CONSTANT)
eclib_error(NODE_SERIAL_RANGE, relnum);
lptr_s[0] = AVAIL_CONSTANT;
lptr_s[1] = lptr[1];
lptr[1] = id;
} /* ZC_delete_row_mem */

ZC_delete_row_file(relnum, id)
short relnum;
long id;
{
short length;
short fd, log_fd;
long top, *lptr;
long nks[2];

id = file_manage(NULL, relnum, 0, 1, id);
length = struct_desc[relnum].size;
fd = struct_desc[relnum].fd;

lock_set(relnum);

top = filesize(fd)/length;
if(id >= top)
eclib_error(NODE_SERIAL_RANGE, relnum);
elseek(fd, id*length, 0);
eread(fd, tupbuf, length);
lptr = (long*) tupbuf;
if(lptr[0] == AVAIL_CONSTANT)
eclib_error(NODE_SERIAL_RANGE, relnum);
elseek(fd, 0L, 0);
eread(fd, ks, 8);
nks[0] = ks[0];
nks[1] = id;
elseek(fd, 0L, 0);
ewrite(fd, nks, 8);
elseek(fd, id*length, 0);
ewrite(fd, ks, 8);
/* reset it, incase we are sequentially scanning through, deleting certain rows */
elseek(fd, (id+1)*length, 0);

index_fixup(tupbuf, tupbuf, id, 0L, relnum);
id += struct_desc[relnum].base;

if(sf_log) {
/* record the change in a log file, in case the disk crashes or something */
log_fd = eopen(logdir, O_BINARY|O_WRONLY|O_APPEND|O_SYNC, 0);
((short *)ks)[0] = relnum;
((short *)ks)[1] = length;
ks[1] = -id;
ewrite(log_fd, ks, 8);
close(log_fd);
}

lock_clear(relnum);
} /* ZC_delete_row_file */

/* insert or update into table */
long ZC_insupd_row_mem(x, id, buf)
struct serial_inmem *x;
long id;
char *buf;
{
char *s, *realloc();
long *lptr, *lptr_s;
short room;
short relnum,  length;
short i;

if(buf == tupbuf) eprint(badbuf);
id = file_manage(x, 0, 1, 1, id);
relnum = x->relnum;
length = struct_desc[relnum].size;

/* no logging or indexes for serial lists in memory */
if(!id) { /* general insert, find an unused serial number */
lptr = (long *) (x->inmem);
if(lptr[0] != AVAIL_CONSTANT)
eclib_error(NODE_CORRUPT, relnum);
if(!struct_desc[relnum].noholes && lptr[1]) {
id = lptr[1];
s = x->inmem + id*length;
lptr_s = (long *)s;
if(lptr_s[0] != AVAIL_CONSTANT)
eclib_error(NODE_CORRUPT, relnum);
lptr[1] = lptr_s[1];
lptr_s[1] = AVAIL_CONSTANT;
} else id = x->n_nodes;
} /* id was 0 */

if(id >= x->n_room) {
room = x->n_room;
room = room * 4 / 3;
if(id >= room) room = id+1;
if(id>>16 || (long)room * length > 65000)
eclib_error(NODE_SERIAL_64K, relnum);

errno = 0;
s = realloc(x->inmem, room*length);
if(!s) eclib_error(NODE_ALLOC_FAILURE, errno);
x->inmem = s;
*(x->inmem_ptr) = s;

s = x->inmem + x->n_room *length;
lptr = (long*) tupbuf;
for(i=0; i<TUPLEN/4; ++i)
*lptr++ = AVAIL_CONSTANT;
while(x->n_room < room) {
memcpy(s, tupbuf, length);
s += length;
++x->n_room;
}
}

if(id >= x->n_nodes) x->n_nodes = id+1;
s = x->inmem + length * id;
lptr = (long*)s;
if(lptr[0] == AVAIL_CONSTANT && lptr[1] != AVAIL_CONSTANT)
eclib_error(NODE_SERIAL_RANGE, relnum);
memcpy(s, buf, length);
return id+x->base;
} /* ZC_insupd_row_mem */

long ZC_insupd_row_file(relnum, id, buf)
short relnum;
long id;
char *buf;
{
long *lptr;
short log_fd, fd;
short length;
char insflag, atend;
long top;

if(buf == tupbuf) eprint(badbuf);
id = file_manage(NULL, relnum, 1, 1, id);
length = struct_desc[relnum].size;
fd = struct_desc[relnum].fd;

lock_set(relnum);
top = filesize(fd)/length;

if(!id) { /* general insert, find an unused serial number */
elseek(fd, 0L, 0);
eread(fd, ks, 8);
if(ks[0] != AVAIL_CONSTANT)
eclib_error(NODE_CORRUPT, relnum);
if(!struct_desc[relnum].noholes && ks[1]) {
id = ks[1];
elseek(fd, id*length, 0);
eread(fd, ks, 8);
if(ks[0] != AVAIL_CONSTANT)
eclib_error(NODE_CORRUPT, relnum);
elseek(fd, -4L, 1);
ewrite(fd, ks, 4);
elseek(fd, 0L, 0);
ewrite(fd, ks, 8);
} else id =top;
} /* id was 0 */

atend = 0;
insflag = 0;

if(id >= top) {
atend = 1;
insflag = 1;
if(id > top) file_extend(fd, id*length, AVAIL_CONSTANT);
}

if(!atend) { /* see if it's a hole */
elseek(fd, id*length, 0);
eread(fd, tupbuf, length);
lptr = (long*)tupbuf;
if(lptr[0] == AVAIL_CONSTANT) {
insflag = 1;
if(lptr[1] != AVAIL_CONSTANT)
eclib_error(NODE_SERIAL_RANGE, relnum);
}
}

elseek(fd, id*length, 0);
ewrite(fd, buf, length);

index_fixup(tupbuf, buf, insflag ? 0L : id, id, relnum);
id += struct_desc[relnum].base;

if(sf_log) {
/* record the change in a log file, in case the disk crashes or something */
log_fd = eopen(logdir, O_BINARY|O_WRONLY|O_APPEND|O_SYNC, 0);
log_append(log_fd, relnum, id, buf);
close(log_fd);
}

lock_clear(relnum);
return id;
} /* ZC_insupd_row_file */

serial_setup(x, relnum, base, n_nodes)
struct serial_inmem *x;
long base;
short relnum, n_nodes;
{
long l;
unsigned int length;
unsigned short i;
extern char **ZC_serial_grabblock();

if(struct_desc[relnum].nodetype != 1)
eclib_error(NODE_STRUCT_TYPE, relnum);
x->setup = 0x5a;
length = struct_desc[relnum].size;

if(n_nodes < 10) n_nodes = 10;
x->n_room = n_nodes;
x->n_nodes = 1;
x->relnum = relnum;
x->base = base;

/* allocate a block at the current level */
l = length * (long) n_nodes;
if(l > 65000)
eclib_error(NODE_SERIAL_64K, relnum);
x->inmem_ptr = ZC_serial_grabblock((int)l, relnum);
x->inmem = *x->inmem_ptr;
for(i=0; i<n_nodes; ++i) {
((long *)(x->inmem + length*i))[0] = AVAIL_CONSTANT;
((long *)(x->inmem + length*i))[1] = AVAIL_CONSTANT;
}
((long *)x->inmem)[1] = 0;
} /* serial_setup */

/* initialize a relation */
serial_rel_create(relnum)
{
int fd;
strcpy(datadir + datadir_len, struct_desc[relnum].name);
fd = eopen(datadir, O_BINARY|O_WRONLY|O_CREAT|O_EXCL, 0666);
ks[0] = AVAIL_CONSTANT;
ks[1] = 0;
ewrite(fd, ks, 8);
file_extend(fd, (long)struct_desc[relnum].size, AVAIL_CONSTANT);
close(fd);
} /* serial_rel_create */

/* create the disk database */
serial_db_create()
{
int i, fd;

if(!sf_dirs)
eclib_error(NODE_SERIAL_DIRS_0,0);

for(i=0; i<ZC_max_struct; ++i) {
if(struct_desc[i].nodetype != 1) continue;
serial_rel_create(i);
} /* end creating data files */

for(i=0; idx_desc[i].cls; ++i) {
indexdir[indexdir_len] = 'a' + i;
fd = eopen(indexdir, O_BINARY|O_WRONLY|O_CREAT|O_EXCL, 0666);
close(fd);
} /* end creating index files */

if(sf_log) {
fd = eopen(logdir, O_BINARY|O_WRONLY|O_CREAT|O_EXCL, 0666);
close(fd);
}
} /* serial_db_create */

/* drop the disk database */
serial_db_drop()
{
int i;

if(!sf_dirs)
eclib_error(NODE_SERIAL_DIRS_0,0);

/* before you remove any files, you better close them,
 * otherwise MSDOS may destroy your file allocation table (bug). */
uncache_fd_all();

for(i=0; i<ZC_max_struct; ++i) {
if(struct_desc[i].nodetype != 1) continue;
strcpy(datadir + datadir_len, struct_desc[i].name);
unlink(datadir);
} /* end removing data files */

for(i=0; idx_desc[i].cls; ++i) {
indexdir[indexdir_len] = 'a' + i;
unlink(indexdir);
} /* end removing index files */

/* logfile not removed; do that manually, if you are really sure */
/* index ratios also not removed. */
} /* serial_db_drop */

serial_mode(multi, log) { sf_multi = multi; sf_log = log; }

/* roll entries forward, read from the logfile */
serial_restore()
{
int fd;
long offset;
short relnum, length;

sf_log = 0; /* cannot log and restore simultaneously */

fd = eopen(logdir, O_BINARY|O_RDONLY, 0);
offset = 0;

while(eread(fd, ks, 8)) {
relnum = ((short *)ks)[0];
length = ((short *)ks)[1];
if((unsigned)relnum >= ZC_max_struct ||
length != struct_desc[relnum].size ||
struct_desc[relnum].nodetype != 1)
eprint("@corrupt logfile %s, relnum %d, length %d, offset %ld",
logdir, relnum, length, offset);

if(ks[1] < 0) /* deletion */
ZC_delete_row_file(relnum, -ks[1]);
else {
eread(fd, logtupbuf, length);
ZC_insupd_row_file(relnum, ks[1], logtupbuf);
offset += length;
}
offset += 8;
} /* until eof */

close(fd);
} /* serial_restore */

ifd_setup(idxnum)
{
short fd;
int mask;

fd = idx_desc[idxnum].fd;
if(fd < 0) {
mask = O_RDONLY|O_BINARY;
if(!(struct_desc[idx_desc[idxnum].relnum].rdonly))
mask |= O_RDWR;
indexdir[indexdir_len] = 'a' + idxnum;
fd = eopen(indexdir, mask, 0);
idx_desc[idxnum].fd = fd;
}
cache_fd(fd, 1, idxnum);
return fd;
} /* ifd_setup */

char *tempidxname()
{
static char newindname[PATHLEN];
strcpy(newindname, indexdir);
newindname[indexdir_len] = '0';
return newindname;
} /* tempidxname */

static fd_setup(relnum)
short relnum;
{
int mask;
short fd;

if(!sf_dirs)
eclib_error(NODE_SERIAL_DIRS_0,0);
if(struct_desc[relnum].nodetype != 1)
eclib_error(NODE_STRUCT_TYPE, relnum);
if(struct_desc[relnum].size > TUPLEN)
eclib_error(NODE_TOO_LARGE, TUPLEN);

fd = struct_desc[relnum].fd;
if(fd < 0) {
mask = O_BINARY|O_RDONLY;
if(!(struct_desc[relnum].rdonly)) mask |= O_RDWR;
strcpy(datadir + datadir_len, struct_desc[relnum].name);
fd = eopen(datadir, mask, 0);
struct_desc[relnum].fd = fd;
}
cache_fd(fd, 0, relnum);
return fd;
} /* fd_setup */

start_scan(c, x, relnum, copy, lowkey, idxname)
struct rel_cursor *c;
struct serial_inmem *x;
char *copy, *idxname;
long lowkey;
short relnum;
{
long offset;
struct IDX_DESC *idxblock;

c->idxnum = -1; /* no index by default */
c->fd = -1;
c->eof = 0;
c->copy = copy;
c->lowkey = lowkey;

if(c->x = x) { /* in memory */
if(x->setup != 0x5a)
eclib_error(NODE_SERIAL_LIST_0, 0);
c->rowid = c->x->base;
c->relnum = x->relnum;
return;
}

if(!copy) c->copy = tupbuf;
c->relnum = relnum;
c->fd = fd_setup(relnum);

if(struct_desc[relnum].holdopen)
eprint("@Accessing relation %s from within a loop over the same relation %s",
struct_desc[relnum].name, struct_desc[relnum].name);

if(!idxname) {
struct_desc[relnum].holdopen = 1;
c->rowid = struct_desc[relnum].base;
return;
}

if(!lowkey) eprint("@null key passed to indexed scan");
c->idxnum = idx_by_name(relnum, idxname);
idxblock = idx_desc + c->idxnum;

lock_set(relnum);
c->ifd = ifd_setup(c->idxnum);
idx_desc[c->idxnum].holdopen = 1;
struct_desc[relnum].holdopen = 0x80 | c->idxnum;

offset = (lowkey * idxblock->numer / idxblock->denom) * 8;
c->eof = 1;
if(offset >= filesize(c->ifd)) return;
c->eof = 0;
elseek(c->ifd, offset, 0);
} /* start_scan */

char *next_scan(c)
struct rel_cursor *c;
{
short length = struct_desc[c->relnum].size;

if(c->eof) return 0;

if(c->idxnum < 0) { /* sequential scan */
if(c->fd >= 0) { /* file */
elseek(c->fd, (c->rowid+1)*length, 0); /* resync file pointer */
while(eread(c->fd, c->copy, length)) {
++c->rowid;
if(*((long *)(c->copy)) != AVAIL_CONSTANT) return c->copy;
}
} else {
while(++c->rowid < c->x->n_nodes) {
char *s = c->x->inmem + length * c->rowid;
if(*((long *)s) != AVAIL_CONSTANT) {
if(c->copy) { memcpy(c->copy, s, length); s = c->copy; }
return s;
}
}
}
c->eof = 1;
return 0;
} /* end sequential */

while(eread(c->ifd, ks, 8)) {
char last = 0;
if(!ks[0]) break; /* hole */
if(ks[0]< 0) last = 1, ks[0] = - ks[0];
if(ks[0] != c->lowkey) continue;

/* if it's last, set eof */
if(last) c->eof = 1;
c->rowid = ks[1] + struct_desc[c->relnum].base;

elseek(c->fd, ks[1]*length, 0);
eread(c->fd, c->copy, length);
return c->copy;
} /* EOF reached */

c->eof = 1;
return 0;
} /* next_scan */

end_scan(c)
struct rel_cursor *c;
{
c->eof = 1;
if(c->fd >= 0) struct_desc[c->relnum].holdopen = 0;
if(c->idxnum >= 0) { /* don't need to hold index open any more */
idx_desc[c->idxnum].holdopen = 0;
lock_clear(c->relnum);
}
} /* end_scan */

static log_append(fd, relnum, id, buf)
long id;
char *buf;
{
short length = struct_desc[relnum].size;
((short *)logtupbuf)[0] = relnum;
((short *)logtupbuf)[1] = length;
((long *)logtupbuf)[1] = id;
memcpy(logtupbuf+8, buf, length);
/* multiple users could append to this logfile simultaneously,
 * if they are updating different relations.
 * I assume Unix write commands are atomic, hence log entries will
 * not collide. */
ewrite(fd, logtupbuf, length+8);
} /* log_append */

rel_unload(fd, relnum)
{
struct rel_cursor c;

start_scan(&c, NULL, relnum, tupbuf, 0L, NULL);
while(next_scan(&c))
log_append(fd, relnum, c.rowid, tupbuf);
end_scan(&c);
} /* rel_unload */

long max_serial(relnum)
{
int fd = fd_setup(relnum);
long l = filesize(fd);
int len = struct_desc[relnum].size;
return struct_desc[relnum].base + l/len;
} /* max_serial */

