/***********************************************************************
|  Interactively configurable forms
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* answer.c: ask a question, get an answer,
 * and perform basic syntax checks on that answer */

#include "icfgram.h"

#define DELIM '\177'

static char line[256];
static struct block *curblock;
static long ask_text;
static short batchnum;
static str_code batchname;
static long rtotal = 0;

/* get next question and answer from the batch file */
nextbatch()
{
int i;
char *s, *t;

if(fgets(line, 255, ansfile) == 0) {
fseek(ansfile, 0L, 2);
inbatch = 0;
return;
}

i = strlen(line);
if(i && line[i-1] == '\n') line[--i] = 0;
if(line[--i] != DELIM)
eprint("line in %s does not end with a proper delimiter", ansfilename);

s = strchr(line,DELIM);
*s++ = 0;
batchnum = atoi(line);
t = strchr(s,DELIM);
*t++ = 0;
batchname = string_to_code(s);
s = strchr(t,DELIM)+1;
t = strchr(s,DELIM);
*t = 0;
strcpy(line, s);
} /* nextbatch */

ask_question(q)
struct block *q;
{
struct block *qi; /* question indirect, refers to another variable */

curblock = q;
qi = q;
if(q->gotonum)
qi = bnum_to_block(q->gotonum);
ask_text = q->ask;
if(is_array(qi) && qi->fmsize && q->nextask)
ask_text = q->nextask;

while(1) { /* keep asking until you get a valid answer */
if(inbatch) {
/* maybe I shouldn't, but I allow mismatched line numbers.
 * Sometimes I want to put in a few extra echo statements or something,
 * and this shifts all the line numbers,
 * yet I want to run from an old session file.
 * So I let this slide.
 * File and questions are still kept in sync by the variable names,
 * which must agree.  If you change a variable name in the form
 * you can always edit the session file and do a global substitute.  But there
 * is no way to deliberately shift the question numbers in a session file.
	if(q->bnum != batchnum)
	eprint("session file %s is answering question %d, but I am asking question %d.",
	ansfilename, batchnum, q->bnum);
*/
if(qi->name != batchname)
eprint("session file %s connects question %d to variable %s, while I assign\n\
question %d to %s.  Perhaps %s was generated from a different form.",
ansfilename, batchnum, code_to_string(batchname),
q->bnum, blockname(qi), ansfilename);
} else {
nl();
output_text(0, ask_text, 0);
printf(": ");
if(gets(line) == 0) exit(0);
spacecrunch(line);
if(!strcmp(line, "?")) {
/* asking for an explanation */
if(q->explain) output_text(0, q->explain, 1);
continue;
}

if(command_check()) continue;
delimcleaner(line);
} /* batch mode test */

if(datatype_check(qi)) {
if(inbatch) nextbatch();
break;
}
if(inbatch)
eprint("value %s in session file %s cannot be interpreted", line, ansfilename);
} /* end while */
} /* ask_question */

/* is this a command, rather than a simple answer to the question? */
command_check()
{
if(line[0] == '!') {
system(line+1);
return 1;
}

if(!strcmp(line, "debug")) {
icfdebug = 1;
return 1;
}

if(!strcmp(line, "quit")) {
tempclean();
puts("goodbye");
exit(0);
}

if(!strncmp(line, "backup", 6)) {
int n;
if(line[6] == 0) backup(1); /* does not return */
if(line[6] != ' ') return 0;
n = atoi(line+7);
backup(n);
}

return 0; /* not a command */
} /* command_check */

static datatype_check(q)
struct block *q;
{
short atype = q->vartype; /* type of answer expected */
short sm = q->stringmod & 0xf;
long l;
char ok, rnext, *s;

/* domains first */
if(atype >= 100) {
l = cv_dom_mm(atype, 0, string_to_code(line), 1);
if(!l) return 0;
} else {

ok = 1;
switch(atype) {
case TYPE_MONEY:
if(line[0] == '$') shiftleft(line);
case TYPE_INT:
rnext = 0;
if(zc_strtol(line, &s, &l)) ok = 0;
if(atype == TYPE_MONEY) {
if(l >= 20000000L) eprint("too many pennies, integer overflow");
l *= 100;
if(*s == '.') {
if(isdigit(*++s)) {
l += (*s-'0')*10;
if(isdigit(*++s)) {
l += *s-'0';
++s;
}
}
}
} /* money */
if(*s == '+') rnext = 1, ++s;
if(*s) ok = 0;
if(!ok) {
printf(atype == 1 ?
"invalid money format, valid formats include $295 or $37.75.\n":
"Invalid number, please enter only digits\n");
return 0;
}
if(rtotal || rnext) {
rtotal += l;
l = rtotal;
printf("Running total = "); printval(stdout, l, atype); nl();
}
if(rnext) return 0;
rtotal = 0;
break;

case TYPE_STRING:
if(sm) smartmix(line, sm-2);
l = string_to_code(line);
break;

case TYPE_BOOL:
smartmix(line, -1);
if(line[1] == 0) {
if(line[0] == 'y') {
l = 1;
break;
}
if(line[0] == 'n') {
l = 0;
break;
}
}
printf("Please enter either Y or N.\n");
return 0;

case TYPE_DATE: /* date */
spacecrunch(line);
l = string_to_date(line, 1900, 2050);
if(!l) {
printf("Invalid date format, enter dates as: 9 29 1937 or 04/16/1984\n");
return 0;
}
break;

default:
eprint("@unexpected type %d", atype);
} /* end switch on type */
} /* end domain or not */

push_val(q, l);
log_qa(q,l);
return 1;
} /* datatype_check */

/* log question and answer */
log_qa(b,ans)
struct block *b;
long ans;
{
if(inbatch) return;
fprintf(ansfile,"%d%c%s%c", curblock->bnum, DELIM, blockname(b), DELIM);
output_text(100, ask_text,0);
fputc(DELIM, ansfile);
printval(ansfile, ans, b->vartype);
fprintf(ansfile, "%c\n", DELIM);
fflush(ansfile);
} /* log_qa */

/* temp file name for stream n */
char *tfname(n)
{
static char s[9];
sprintf(s, "icftmp%02d", n);
return s;
} /* tfname */

output_text(stream, offset, nl)
int stream; /* which output file */
long offset; /* location in file where text begins */
{
FILE *outf; /* output file, could be stdout */
char ident[41]; /* identifier */
int i, c, atmode = 0;
long varval;
short vartype;

if(stream == 0) {
if(inbatch) return; /* supress printing */
outf = stdout;
} else if(stream < 100) {
outf = efopen(tfname(stream), streamuse[stream] ? "a" : "w");
streamuse[stream] = 1;
} else {
/* streams over 100 copy questions to session files for future review and/or edit */
outf = ansfile;
}

fseek(formfile, offset, 0);

while((c = getc(formfile)) != EOF) {
if(atmode) { /* gathering ident */
if(!isalnum(c) && c != '_') {
/* end of ident */
ident[i] = 0;
var_value(string_to_code(ident), &varval, &vartype);
printval(outf, varval, vartype);
atmode = 0;
goto top; /* process c */
} /* not part of ident */
if(i < 40) ident[i++] = c;
continue;
} /* end atmode */
top:
if(c == '`') break;
if(c == '@') {
atmode = 1;
i = 0;
continue;
}
if(stream >= 100 && c == '\n') break;
putc(c, outf);
} /* while copying text */

if(nl) fprintf(outf, "\n");
if(stream && stream < 100) fclose(outf);
} /* output_text */

/* because we store arrays of strings in a `unload' file as pipe delimited
 * fields, like sql, we don't allow pipes in the data itself;
 * neither should you use pipes in the questions you ask the user. */
delimcleaner(s)
char *s;
{
for(; *s; ++s)
if(*s == DELIM) *s = '\\';
} /* delimcleaner */

/* oops, back up n questions */
backup(n)
int n;
{
int cnt, i;
FILE *f;
struct block *b;

rtotal = 0;

/* count lines in existing file */
rewind(ansfile);
cnt = 0;
while(fgets(line, 250, ansfile))
++cnt;
cnt -= n;
if(cnt < 0) cnt = 0;

rewind(ansfile);
f = efopen("icftmpbk", "w");
for(i=0; i<cnt; ++i) {
fgets(line, 250, ansfile);
fputs(line, f);
}
fclose(f);
fclose(ansfile);
erename("icftmpbk", ansfilename);
ansfile = efopen(ansfilename, "r+");
inbatch = 1;
nextbatch();

/* null out all variables */
b = bnum_to_block(1);
for(i=1; i<blist.n_nodes; ++i, ++b)
b->notnull = 0;
tempclean();

/* jump to block 1 and we're done */
longjmp(actjmp, 1);
} /* backup */

tempclean()
{
int i;

for(i=1; i<100; ++i)
if(streamuse[i]) {
unlink(tfname(i));
streamuse[i] = 0;
}
} /* tempclean */

